#include <math.h>
#include "profile.h"

/**
 * SECTION:profile
 * @Short_description: A profile histogram.
 *
 * A object to study the dependency of one variable with respect to
 * other by storing the mean value of the former as a function
 * of the later.
 *
 */

static void       prof_class_init          (HstProfileClass *klass);
static void       prof_init                (HstProfile      *object,
                                            HstProfileClass *klass);
static void       prof_dispose             (GObject        *object);
static void       prof_finalize            (GObject        *object);
static void       prof_get_property        (GObject       *object,
                                           guint           property_id,
                                           GValue         *value,
                                           GParamSpec     *pspec);
static void       prof_fill(HstHistogram *hst, double x, double w);
static void       prof_reset(HstHistogram *hst);
static void       prof_range_changed(HstFunction *f, gpointer p);
static gpointer   parent_class = NULL;


enum {
    PROP_0,
    PROP_RMS,
    PROP_NTOT
} ;

GType hst_profile_get_type (void)
{
    static GType object_type = 0;

    if (!object_type)
    {
        static const GTypeInfo object_info = {
            sizeof (HstProfileClass),
            (GBaseInitFunc) NULL,
            (GBaseFinalizeFunc) NULL,
            (GClassInitFunc) prof_class_init,
            NULL,               /* class_finalize */
            NULL,               /* class_data */
            sizeof (HstProfile),
            16,         /* n_preallocs */
            (GInstanceInitFunc) prof_init,
            NULL,               /* value_table */
        };

        object_type = g_type_register_static (HST_TYPE_HISTOGRAM,
                                              "HstProfile",
                                              &object_info, 0);
    }

    return object_type;
}

static void
prof_class_init (HstProfileClass *class)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (class);
    HstHistogramClass *fclass = HST_HISTOGRAM_CLASS(class);
    parent_class = g_type_class_ref (HST_TYPE_HISTOGRAM);

    gobject_class->get_property = prof_get_property;
    gobject_class->dispose = prof_dispose;
    gobject_class->finalize = prof_finalize;

    fclass->fill  = prof_fill;
    fclass->reset = prof_reset;


    /**
     * rms
     *
     * Retrieves the pointer of the array with the
     * RMS of the histogram contents.
     */
    g_object_class_install_property (
        gobject_class,
        PROP_RMS,
        g_param_spec_pointer ("rms", "RMS of profile data",
                              "Pointer to rms of data",
                              G_PARAM_READABLE)
        );

    /**
     * ntotal
     *
     * Retrieves the pointer of the array with the
     * number of events used to compute the profile data
     * (mean value) and RMS.
     */
    g_object_class_install_property (
        gobject_class,
        PROP_NTOT,
        g_param_spec_pointer ("ntotal", "events of profile data",
                              "Pointer to statistics of profile data",
                              G_PARAM_READABLE)
        );

}

static void
prof_init( HstProfile *h, HstProfileClass *klass)
{
    int i;
    HstFunction *f;
    g_return_if_fail( HST_IS_PROFILE(h) );

    f = HST_FUNCTION(h);
    h->n = g_new(double,f->nbin);
    h->s = g_new(double,f->nbin);
    for (i=0;i<f->nbin;i++)
    {
        h->n[i] = 0.;
        h->s[i] = 0.;
    }
    g_signal_connect(f,"range-changed",
                     G_CALLBACK(prof_range_changed),
                     NULL
        );
}

static void
prof_get_property (GObject     *object,
                   guint        property_id,
                   GValue      *value,
                   GParamSpec  *pspec)
{
    HstProfile *h;
    g_return_if_fail( HST_IS_PROFILE(object) );

    h = HST_PROFILE(object);

    switch (property_id)
    {
    case PROP_RMS:
    {
        int ib;
        int nbin = h->H.parent_instance.nbin;
        double *ss = g_new0(double, nbin);
        for (ib=0; ib<nbin; ib++)
        {
            double mean = h->H.y[ib];
            ss[ib] = sqrt( fabs(h->s[ib] - mean*mean) );
        }
        g_value_set_pointer (value,(void *)ss);
        break;
    }
    case PROP_NTOT:
        g_value_set_pointer(value, (void *)h->n);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
        break;
    }
}

static void
prof_range_changed(HstFunction *f, gpointer p)
{
    int i;
    HstProfile  *h = HST_PROFILE(f);
    g_free( h->n );
    g_free( h->s );
    h->n = g_new(double,f->nbin);
    h->s = g_new(double,f->nbin);
    for (i=0;i<f->nbin;i++) {
        h->n[i] = 0.;
        h->s[i] = 0.;
    }
}

/**
 * hst_profile_new:
 * @n: number of bins in the histogram
 * @xmin: lower range of the X axis
 * @xmax: upper range of the X axis
 * @ptr: a pointer to an array of doubles. HstHistogram will use it
 *       internally as the contents of the histogram. It can be
 *       used to monitor a given array or to retrieve or fill the
 *       buffer with the contents of the histogram
 *
 * This is a convenience function to create a #HstProfile and
 * setting its properties in one go.
 *
 * Return value: a HstProfile pointer
 *
 */
HstProfile *
hst_profile_new(int n, double xmin, double xmax, double *ptr)
{
    HstProfile *h;
    g_return_val_if_fail ( n>0, NULL );
    h = g_object_new (HST_TYPE_PROFILE, "nx",n,"xmin",xmin,"xmax",xmax,"ptr",ptr,NULL);
    hst_histogram_reset(HST_HISTOGRAM(h));
    return h;
}

static void
prof_dispose(GObject *object)
{
    G_OBJECT_CLASS (parent_class)->dispose (object);
}

static void
prof_finalize (GObject *object)
{
    HstProfile *hst = HST_PROFILE(object);
    g_free(hst->n);
    g_free(hst->s);


    G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
prof_reset(HstHistogram *hst)
{
    int i;
    HstFunction *f = HST_FUNCTION(hst);
    HstProfile  *h = HST_PROFILE(hst);
    for (i=0;i<f->nbin;i++)
    {
        h->n[i] = 0.;
        h->s[i] = 0.;
    }
    HST_HISTOGRAM_CLASS (parent_class)->reset (hst);
}

static void
prof_fill(HstHistogram *h, double x, double w)
{
    hst_profile_fill(HST_PROFILE(h), x, w, 1.);
}

/**
 * hst_profile_fill:
 * @hst:  pointer of a #HstProfile object
 * @x:    value for X
 * @y:    value for Y that will be averaged
 * @w:    weight
 *
 */
void
hst_profile_fill(HstProfile *hst, double x, double y, double w)
{
    HstFunction  *f;
    HstHistogram *h;
    int wbin;
    double np;

    g_return_if_fail( HST_IS_PROFILE(hst) );
    f = HST_FUNCTION(hst);
    h = HST_HISTOGRAM(hst);
    wbin = hst_function_find_bin(f,x);

    if ( wbin<0 || wbin>=f->nbin )
        return;

    np = hst->n[wbin] + w;
    h->y[wbin] = (hst->n[wbin] * h->y[wbin] + w*y)/np;
    hst->s[wbin] = (hst->n[wbin] * hst->s[wbin] + w*y*y)/np;
    hst->n[wbin] = np;

    hst_stats_add(h->stat,x,y,w);

    if ( x> h->vmax )
        h->vmax = x;
    if ( x < h->vmin )
        h->vmin = x;
}
